@@ -6,6 +6,8 @@
 #[macro_use]
 mod tests;
 
+use itertools::Itertools;
+use std::cmp::Ordering;
 use std::collections::HashSet;
 use std::fmt::*;
 use std::rc::Rc;
@@ -38,6 +40,20 @@ pub struct Location {
     pub col: usize,
 }
 
+impl Ord for Location {
+    fn cmp(&self, other: &Self) -> Ordering {
+        self.row
+            .cmp(&other.row)
+            .then_with(|| self.col.cmp(&other.col))
+    }
+}
+
+impl PartialOrd for Location {
+    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
+        Some(self.cmp(other))
+    }
+}
+
 /// Records the result of the parser.
 ///
 /// Similar to the [`std::result::Result`](https://doc.rust-lang.org/std/result/enum.Result.html) type, outputs `ParseResult::Ok` if parsing succeeds
@@ -249,12 +265,47 @@ impl<'a, T, S: Clone> ParseResult<'a, T, S> {
             },
         }
     }
+
+    pub fn backtrackable(self) -> ParseResult<'a, T, S> {
+        match self {
+            ParseResult::Ok {
+                input,
+                location,
+                output,
+                state,
+                ..
+            } => ParseResult::Ok {
+                input,
+                location,
+                output,
+                state,
+                committed: false,
+            },
+            ParseResult::Err {
+                message,
+                from,
+                to,
+                state,
+                ..
+            } => ParseResult::Err {
+                message: message,
+                from,
+                to,
+                state,
+                committed: false,
+            },
+        }
+    }
+
     fn unwrap(self, source: &'a str) -> T {
         match self {
             ParseResult::Ok { output, .. } => output,
             ParseResult::Err {
                 message, from, to, ..
-            } => panic!("{}", display_error(source, message, from, to)),
+            } => {
+                println!("{}", display_error(source, message, from, to));
+                panic!();
+            }
         }
     }
     fn unwrap_err(self) -> String
@@ -476,6 +527,15 @@ pub trait Parser<'a, Output, State: Clone> {
     {
         BoxedParser::new(left(self, ignored_parser))
     }
+
+    fn backtrackable(self) -> BoxedParser<'a, Output, State>
+    where
+        Self: Sized + 'a,
+        Output: Clone + 'a,
+        State: 'a,
+    {
+        BoxedParser::new(backtrackable(self))
+    }
 }
 
 impl<'a, F, Output, State: 'a> Parser<'a, Output, State> for F
@@ -561,34 +621,36 @@ where
 }
 
 /// Parse a given token string.
+///
+/// ⚠️ Newlines are not allowed in tokens.
 pub fn token<'a, S: Clone + 'a>(expected: &'static str) -> BoxedParser<'a, &str, S> {
     BoxedParser::new(move |input: &'a str, location: Location, state: S| {
-        let found = input.get(0..expected.len());
-        match found {
-            Some(next) if next == expected => ParseResult::Ok {
+        let peek = input.get(0..expected.len());
+        if peek == Some(expected) {
+            ParseResult::Ok {
                 input: &input[expected.len()..],
                 output: expected,
-                location: increment_col(expected.len(), location),
+                location: increment_col(unicode_column_width(expected), location),
                 state,
                 committed: true,
-            },
-            _ => ParseResult::Err {
+            }
+        } else {
+            let expected_len = expected.graphemes(true).count();
+            let peek_str = &input.graphemes(true).take(expected_len).join("");
+            ParseResult::Err {
                 message: format!(
                     "I'm expecting a `{}` but found {}.",
                     expected,
-                    display_token(found)
+                    display_token(peek_str)
                 ),
                 from: location,
-                to: match found {
-                    Some(found_str) => Location {
-                        col: location.col + found_str.len(),
-                        ..location
-                    },
-                    None => location,
+                to: Location {
+                    col: location.col + unicode_column_width(peek_str),
+                    ..location
                 },
                 state,
                 committed: false,
-            },
+            }
         }
     })
 }
@@ -607,10 +669,12 @@ fn increment_row(additional_row: usize, location: Location) -> Location {
     }
 }
 
-fn display_token<T: Display>(token: Option<T>) -> String {
-    match token {
-        Some(token_content) => format!("`{}`", token_content).replace("\n", "\\n"),
-        None => "nothing".to_string(),
+fn display_token<T: Display>(token: T) -> String {
+    let token_str = &token.to_string();
+    if unicode_column_width(token_str) == 0 {
+        "nothing".to_string()
+    } else {
+        format!("`{}`", token_str.replace("\n", "\\n"))
     }
 }
 
@@ -696,6 +760,13 @@ where
     }
 }
 
+fn backtrackable<'a, P: 'a, A, S: Clone + 'a>(parser: P) -> impl Parser<'a, A, S>
+where
+    P: Parser<'a, A, S>,
+{
+    move |input, location, state| parser.parse(input, location, state).backtrackable()
+}
+
 /// Run the left parser, then the right, last keep the left result and discard the right.
 fn left<'a, P1: 'a, P2: 'a, R1: 'a, R2: 'a, S: Clone + 'a>(
     parser1: P1,
@@ -813,9 +884,8 @@ where
                 from,
                 to,
                 state,
-                committed: cur_committed,
+                committed,
             } => {
-                committed |= cur_committed;
                 return ParseResult::Err {
                     message,
                     from,
@@ -843,11 +913,23 @@ where
                     output.push(cur_item);
                 }
                 ParseResult::Err {
-                    committed: cur_committed,
-                    ..
+                    message,
+                    from,
+                    to,
+                    state,
+                    committed,
                 } => {
-                    committed |= cur_committed;
-                    break;
+                    if committed {
+                        return ParseResult::Err {
+                            message,
+                            from,
+                            to,
+                            state,
+                            committed,
+                        };
+                    } else {
+                        break;
+                    }
                 }
             }
         }
@@ -864,7 +946,9 @@ where
 
 /// Run the parser zero or more times until an end delimiter (or end of input) and combine each output into a vector of outputs.
 pub fn zero_or_more_until<'a, P, A, S: Clone + 'a, E, B>(
+    item_name: &'a str,
     parser: P,
+    end_name: &'a str,
     end_parser: E,
 ) -> impl Parser<'a, Vec<A>, S>
 where
@@ -892,10 +976,21 @@ where
                 };
             }
             ParseResult::Err {
-                committed: cur_committed,
+                message,
+                from,
+                to,
+                committed,
                 ..
             } => {
-                committed |= cur_committed;
+                if committed {
+                    return ParseResult::Err {
+                        message,
+                        from,
+                        to,
+                        state,
+                        committed,
+                    };
+                }
             } // no big deal, continue parsing
         }
 
@@ -960,10 +1055,21 @@ where
                     };
                 }
                 ParseResult::Err {
-                    committed: cur_committed,
+                    message,
+                    from,
+                    to,
+                    committed: err_committed,
                     ..
                 } => {
-                    committed |= cur_committed;
+                    if err_committed {
+                        return ParseResult::Err {
+                            message,
+                            from,
+                            to,
+                            state,
+                            committed: err_committed,
+                        };
+                    }
                     match parser.parse(input, location, state.clone()) {
                         ParseResult::Ok {
                             input: cur_input,
@@ -996,8 +1102,7 @@ where
                                 };
                             } else {
                                 return ParseResult::Err {
-                                    message: "I'm expecting either the intended string or the end delimiter. However, neither was found."
-                                        .to_string(),
+                                    message: format!("I'm expecting either {} or {}. However, neither was found.", item_name, end_name),
                                     from: end_from,
                                     to: end_to,
                                     state,
@@ -1014,7 +1119,9 @@ where
 
 /// Run the parser one or more times until an end delimiter (or end of input) and combine each output into a vector of outputs.
 pub fn one_or_more_until<'a, P, A, S: Clone + 'a, E, B>(
+    item_name: &'a str,
     parser: P,
+    end_name: &'a str,
     end_parser: E,
 ) -> impl Parser<'a, Vec<A>, S>
 where
@@ -1033,8 +1140,10 @@ where
             } => {
                 committed |= cur_committed;
                 return ParseResult::Err {
-                    message: "I'm expecting at least one occurrence of the intended string but reached the end delimiter."
-                        .to_string(),
+                    message: format!(
+                        "I'm expecting at least one occurrence of {} but reached {}.",
+                        item_name, end_name
+                    ),
                     from: location,
                     to: end_location,
                     state,
@@ -1042,10 +1151,21 @@ where
                 };
             }
             ParseResult::Err {
-                committed: cur_committed,
+                message,
+                from,
+                to,
+                committed: err_committed,
                 ..
             } => {
-                committed |= cur_committed;
+                if err_committed {
+                    return ParseResult::Err {
+                        message,
+                        from,
+                        to,
+                        state,
+                        committed: err_committed,
+                    };
+                }
                 match parser.parse(input, location, state.clone()) {
                     ParseResult::Ok {
                         input: cur_input,
@@ -1091,48 +1211,64 @@ where
                         committed,
                     };
                 }
-                ParseResult::Err { .. } => match parser.parse(input, location, state.clone()) {
-                    ParseResult::Ok {
-                        input: cur_input,
-                        output: cur_item,
-                        location: cur_location,
-                        state: cur_state,
-                        committed: cur_committed,
-                    } => {
-                        input = cur_input;
-                        location = cur_location;
-                        state = cur_state;
-                        committed |= cur_committed;
-                        output.push(cur_item);
+                ParseResult::Err {
+                    message,
+                    from,
+                    to,
+                    committed: err_committed,
+                    ..
+                } => {
+                    if err_committed {
+                        return ParseResult::Err {
+                            message,
+                            from,
+                            to,
+                            state,
+                            committed: err_committed,
+                        };
                     }
-                    ParseResult::Err {
-                        from: end_from,
-                        to: end_to,
-                        committed: cur_committed,
-                        ..
-                    } => {
-                        committed |= cur_committed;
-                        if input == "" {
-                            // reached the end of input
-                            return ParseResult::Ok {
-                                input,
-                                output,
-                                location,
-                                state,
-                                committed,
-                            };
-                        } else {
-                            return ParseResult::Err {
-                                message: "I'm expecting either the intended string or the end delimiter. However, neither was found."
-                                    .to_string(),
+                    match parser.parse(input, location, state.clone()) {
+                        ParseResult::Ok {
+                            input: cur_input,
+                            output: cur_item,
+                            location: cur_location,
+                            state: cur_state,
+                            committed: cur_committed,
+                        } => {
+                            input = cur_input;
+                            location = cur_location;
+                            state = cur_state;
+                            committed |= cur_committed;
+                            output.push(cur_item);
+                        }
+                        ParseResult::Err {
+                            from: end_from,
+                            to: end_to,
+                            committed: cur_committed,
+                            ..
+                        } => {
+                            committed |= cur_committed;
+                            if input == "" {
+                                // reached the end of input
+                                return ParseResult::Ok {
+                                    input,
+                                    output,
+                                    location,
+                                    state,
+                                    committed,
+                                };
+                            } else {
+                                return ParseResult::Err {
+                                message: format!("I'm expecting either {} or {}. However, neither was found.", item_name, end_name),
                                 from: end_from,
                                 to: end_to,
                                 state,
                                 committed,
                             };
+                            }
                         }
                     }
-                },
+                }
             }
         }
     }
@@ -1172,19 +1308,41 @@ where
         let mut output = Vec::new();
         let mut committed = false;
 
-        while let ParseResult::Ok {
-            input: cur_input,
-            output: cur_item,
-            location: cur_location,
-            state: cur_state,
-            committed: cur_committed,
-        } = parser.parse(input, location, state.clone())
-        {
-            input = cur_input;
-            location = cur_location;
-            state = cur_state;
-            committed |= cur_committed;
-            output.push(cur_item);
+        loop {
+            match parser.parse(input, location, state.clone()) {
+                ParseResult::Ok {
+                    input: cur_input,
+                    output: cur_item,
+                    location: cur_location,
+                    state: cur_state,
+                    committed: cur_committed,
+                } => {
+                    input = cur_input;
+                    location = cur_location;
+                    state = cur_state;
+                    committed |= cur_committed;
+                    output.push(cur_item);
+                }
+                ParseResult::Err {
+                    message,
+                    from,
+                    to,
+                    state,
+                    committed,
+                } => {
+                    if committed {
+                        return ParseResult::Err {
+                            message,
+                            from,
+                            to,
+                            state,
+                            committed,
+                        };
+                    } else {
+                        break;
+                    }
+                }
+            }
         }
 
         ParseResult::Ok {
@@ -1426,7 +1584,7 @@ where
                     message: format!(
                         "I'm expecting {} but found {}.",
                         expecting,
-                        display_token(Some(content)),
+                        display_token(content),
                     )
                     .to_string(),
                     from: location,
@@ -1979,13 +2137,14 @@ pub fn sequence<'a, A: Clone + 'a, S: Clone + 'a>(
             vec![],
             pair(
                 item.clone(),
-                left(
-                    zero_or_more(right(
-                        wrap(spaces.clone(), token(separator), spaces.clone()),
-                        item.clone(),
-                    )),
-                    right(
-                        spaces.clone(),
+                right(
+                    spaces.clone(),
+                    left(
+                        zero_or_more(wrap(
+                            left(token(separator), spaces.clone()).backtrackable(),
+                            item.clone(),
+                            spaces.clone(),
+                        )),
                         match trailing {
                             Trailing::Forbidden => token(""),
                             Trailing::Optional => {
@@ -2082,8 +2241,10 @@ pub fn display_error(source: &str, error_message: String, from: Location, to: Lo
     } else {
         to.col - from.col
     };
-    let error_line = row.to_string() + "| " + source.split("\n").collect::<Vec<&str>>()[row - 1];
-    let error_pointer = " ".repeat(col - 1 + row.to_string().len() + 2) + &"^".repeat(error_length);
+    let row_tag = row.to_string();
+    let row_tag_len = row_tag.len();
+    let error_line = row_tag + "| " + source.split("\n").collect::<Vec<&str>>()[row - 1];
+    let error_pointer = " ".repeat(col - 1 + row_tag_len + 2) + &"^".repeat(error_length);
     let error_report = error_line + "\n" + &error_pointer + "\n" + "⚠️ " + &error_message;
     error_report
 }
