@@ -101,7 +101,50 @@ fn test_update() {
 
 #[test]
 fn test_token() {
-    assert_succeed(token(""), "aweiow", "");
+    assert_succeed(token(""), "abc", "");
+    assert_succeed(token("a"), "abc", "a");
+    assert_succeed(token("ab"), "abc", "ab");
+    assert_succeed(token("abc"), "abc", "abc");
+
+    // Normalization Form C (NFC) Characters are decomposed and then re-composed by canonical equivalence
+    assert_succeed(token("ãáç"), "ãáç", "ãáç");
+    // Normalization Form D (NFD) Characters are decomposed by canonical equivalence
+    assert_succeed(token("ãáç"), "ãáç", "ãáç");
+
+    // Normalization Form C (NFC) Characters are decomposed and then re-composed by canonical equivalence
+    assert_fail(token("ãáç"), "ãá", "I'm expecting a `ãáç` but found `ãá`.");
+    // Normalization Form D (NFD) Characters are decomposed by canonical equivalence
+    assert_fail(token("ãáç"), "ãá", "I'm expecting a `ãáç` but found `ãá`.");
+
+    // Normalization Form C (NFC) Characters are decomposed and then re-composed by canonical equivalence
+    assert_fail(
+        token("ãáç"),
+        "ãá\nabc",
+        "I'm expecting a `ãáç` but found `ãá\\n`.",
+    );
+    // Normalization Form D (NFD) Characters are decomposed by canonical equivalence
+    assert_fail(
+        token("ãáç"),
+        "ãá\nabc",
+        "I'm expecting a `ãáç` but found `ãá\\n`.",
+    );
+
+    assert_succeed(
+        token("👩‍🔬👩🏿‍💻我哋👨‍👨‍👧‍👧呢一餐"),
+        "👩‍🔬👩🏿‍💻我哋👨‍👨‍👧‍👧呢一餐",
+        "👩‍🔬👩🏿‍💻我哋👨‍👨‍👧‍👧呢一餐",
+    );
+    assert_succeed(
+        token("👩‍🔬👩🏿‍💻"),
+        "👩‍🔬👩🏿‍💻我哋👨‍👨‍👧‍👧呢一餐",
+        "👩‍🔬👩🏿‍💻",
+    );
+
+    assert_fail(
+        token("👩‍🔬👩🏿‍💻我哋👨‍👨‍👧‍👧呢一餐"),
+        "👩‍🔬👩🏿‍💻我哋👨‍👨‍👧‍👧呢",
+        "I'm expecting a `👩‍🔬👩🏿‍💻我哋👨‍👨‍👧‍👧呢一餐` but found `👩‍🔬👩🏿‍💻我哋👨‍👨‍👧‍👧呢`.",
+    );
 }
 
 #[test]
@@ -189,6 +232,64 @@ fn test_one_or_more() {
         "bbc",
         "I'm expecting a `a` but found `b`.",
     );
+
+    assert_succeed(
+        succeed!(identity).keep(one_of!(
+            succeed!(identity)
+                .keep(one_or_more(token("a")))
+                .skip(token("b")),
+            token("aab").map(|s| vec![s])
+        )),
+        "aab",
+        vec!["a", "a"],
+    );
+
+    assert_fail(
+        succeed!(identity).keep(one_of!(
+            succeed!(identity)
+                .keep(one_or_more(token("a")))
+                .skip(token("b")),
+            token("aab").map(|s| vec![s])
+        )),
+        "b",
+        "I'm expecting a `aab` but found `b`.",
+    );
+
+    assert_fail(
+        succeed!(identity).keep(one_of!(
+            succeed!(identity)
+                .keep(one_or_more(token("aa")))
+                .skip(token("!")),
+            token("aab").map(|s| vec![s])
+        )),
+        "aab",
+        "I'm expecting a `!` but found `b`.",
+    );
+
+    assert_fail(
+        succeed!(identity).keep(one_of!(
+            succeed!(identity)
+                .keep(one_or_more(
+                    succeed!(identity).keep(token("aa")).skip(token("!"))
+                ))
+                .skip(token("?")),
+            token("aab").map(|s| vec![s])
+        )),
+        "aab",
+        "I'm expecting a `!` but found `b`.",
+    );
+    assert_fail(
+        succeed!(identity).keep(one_of!(
+            succeed!(identity)
+                .keep(one_or_more(
+                    succeed!(identity).keep(token("aa")).skip(token("!"))
+                ))
+                .skip(token("?")),
+            token("aab").map(|s| vec![s])
+        )),
+        "aa!aab",
+        "I'm expecting a `!` but found `b`.",
+    );
 }
 
 #[test]
@@ -442,36 +543,63 @@ fn test_optional_with_default() {
 #[test]
 fn test_one_or_more_until() {
     assert_succeed(
-        succeed!(identity).keep(one_or_more_until(token("a"), token("b"))),
+        succeed!(identity).keep(one_or_more_until(
+            "the letter `a`",
+            token("a"),
+            "the ending `b`",
+            token("b"),
+        )),
         "ab",
         vec!["a"],
     );
     assert_succeed(
-        succeed!(identity).keep(one_or_more_until(token("a"), token("b"))),
+        succeed!(identity).keep(one_or_more_until(
+            "the letter `a`",
+            token("a"),
+            "the ending `b`",
+            token("b"),
+        )),
         "aaab",
         vec!["a", "a", "a"],
     );
     assert_fail(
-        succeed!(identity).keep(one_or_more_until(token("a"), token("b"))),
+        succeed!(identity).keep(one_or_more_until(
+            "the letter `a`",
+            token("a"),
+            "the ending `b`",
+            token("b"),
+        )),
         "b",
-        "I'm expecting at least one occurrence of the intended string but reached the end delimiter."
+        "I'm expecting at least one occurrence of the letter `a` but reached the ending `b`.",
     );
     assert_fail(
-        succeed!(identity).keep(one_or_more_until(token("a"), token("b"))),
+        succeed!(identity).keep(one_or_more_until(
+            "the letter `a`",
+            token("a"),
+            "the ending `b`",
+            token("b"),
+        )),
         "aaac",
-        "I'm expecting either the intended string or the end delimiter. However, neither was found.",
+        "I'm expecting either the letter `a` or the ending `b`. However, neither was found.",
     );
     assert_succeed(
-        succeed!(identity).keep(one_or_more_until(token("a"), token("b"))),
+        succeed!(identity).keep(one_or_more_until(
+            "the letter `a`",
+            token("a"),
+            "the ending `b`",
+            token("b"),
+        )),
         "aaa",
         vec!["a", "a", "a"],
     );
 
     assert_succeed(
         succeed!(identity).keep(one_or_more_until(
+            "the line",
             succeed!(|line| line)
                 .keep(take_chomped(chomp_while0c(|c| *c != '\n', "line")))
                 .skip(token("\n")),
+            "the end symbol",
             token("<end>"),
         )),
         "this is the 1st line
@@ -491,9 +619,11 @@ this is the 4th line
 
     assert_succeed(
         succeed!(identity).keep(one_or_more_until(
+            "the line",
             succeed!(|line| line)
                 .keep(take_chomped(chomp_while0c(|c| *c != '\n', "line")))
                 .skip(token("\n")),
+            "the end symbol",
             token("<end>"),
         )),
         "this is the 1st line
@@ -511,54 +641,236 @@ this is the 4th line
 
     assert_fail(
         succeed!(identity).keep(one_or_more_until(
+            "the line",
             succeed!(|line| line)
                 .keep(take_chomped(chomp_while0c(|c| *c != '\n', "line")))
                 .skip(token("\n")),
+            "the end symbol",
             token("<end>"),
         )),
         "<end>",
-        "I'm expecting at least one occurrence of the intended string but reached the end delimiter.",
+        "I'm expecting at least one occurrence of the line but reached the end symbol.",
+    );
+
+    // adapted from test_zero_or_more_until()
+
+    assert_fail(
+        succeed!(identity).keep(one_of!(
+            succeed!(identity)
+                .keep(one_or_more_until(
+                    "a letter `a`",
+                    token("a"),
+                    "an end symbol",
+                    token("^")
+                ))
+                .skip(token("b")),
+            token("aab").map(|s| vec![s])
+        )),
+        "aab",
+        "I'm expecting either a letter `a` or an end symbol. However, neither was found.",
+    );
+
+    assert_fail(
+        succeed!(identity).keep(one_of!(
+            succeed!(identity)
+                .keep(one_or_more_until(
+                    "a letter `a`",
+                    token("a"),
+                    "an end symbol",
+                    token("^")
+                ))
+                .skip(token("b")),
+            token("aab").map(|s| vec![s])
+        )),
+        "b",
+        "I'm expecting a `aab` but found `b`.",
+    );
+
+    assert_fail(
+        succeed!(identity).keep(one_of!(
+            succeed!(identity)
+                .keep(one_or_more_until(
+                    "an `aa`",
+                    token("aa"),
+                    "an end symbol",
+                    token("^")
+                ))
+                .skip(token("!")),
+            token("aab").map(|s| vec![s])
+        )),
+        "aab",
+        "I'm expecting either an `aa` or an end symbol. However, neither was found.",
+    );
+
+    assert_fail(
+        succeed!(identity).keep(one_of!(
+            succeed!(identity)
+                .keep(one_or_more_until(
+                    "an `aa!`",
+                    succeed!(identity).keep(token("aa")).skip(token("!")),
+                    "an end symbol",
+                    token("^")
+                ))
+                .skip(token("?")),
+            token("aab").map(|s| vec![s])
+        )),
+        "aab",
+        "I'm expecting a `!` but found `b`.",
+    );
+
+    // committed err when running end_parser
+    assert_fail(
+        succeed!(identity).keep(one_of!(
+            succeed!(identity)
+                .keep(one_or_more_until(
+                    "an `aa`",
+                    token("aa"),
+                    "an end symbol",
+                    succeed!(identity).keep(token("b")).skip(token("!"))
+                ))
+                .skip(token("?")),
+            token("aab").map(|s| vec![s])
+        )),
+        "aab",
+        "I'm expecting a `!` but found nothing.",
+    );
+
+    assert_fail(
+        succeed!(identity).keep(one_of!(
+            succeed!(identity)
+                .keep(one_or_more_until(
+                    "an `aa`",
+                    token("aa"),
+                    "an end symbol",
+                    succeed!(identity).keep(token("b")).skip(token("!"))
+                ))
+                .skip(token("?")),
+            token("aab").map(|s| vec![s])
+        )),
+        "b",
+        "I'm expecting a `!` but found nothing.",
+    );
+}
+
+#[test]
+fn test_zero_or_more() {
+    assert_succeed(
+        succeed!(identity).keep(one_of!(
+            succeed!(identity)
+                .keep(zero_or_more(token("a")))
+                .skip(token("b")),
+            token("aab").map(|s| vec![s])
+        )),
+        "aab",
+        vec!["a", "a"],
+    );
+
+    assert_succeed(
+        succeed!(identity).keep(one_of!(
+            succeed!(identity)
+                .keep(zero_or_more(token("a")))
+                .skip(token("b")),
+            token("aab").map(|s| vec![s])
+        )),
+        "b",
+        vec![],
+    );
+
+    assert_fail(
+        succeed!(identity).keep(one_of!(
+            succeed!(identity)
+                .keep(zero_or_more(token("aa")))
+                .skip(token("!")),
+            token("aab").map(|s| vec![s])
+        )),
+        "aab",
+        "I'm expecting a `!` but found `b`.",
+    );
+
+    assert_fail(
+        succeed!(identity).keep(one_of!(
+            succeed!(identity)
+                .keep(zero_or_more(
+                    succeed!(identity).keep(token("aa")).skip(token("!"))
+                ))
+                .skip(token("?")),
+            token("aab").map(|s| vec![s])
+        )),
+        "aab",
+        "I'm expecting a `!` but found `b`.",
     );
 }
 
 #[test]
 fn test_zero_or_more_until() {
     assert_succeed(
-        succeed!(identity).keep(zero_or_more_until(token("a"), token("b"))),
+        succeed!(identity).keep(zero_or_more_until(
+            "a letter `a`",
+            token("a"),
+            "an ending `b`",
+            token("b"),
+        )),
         "ab",
         vec!["a"],
     );
     assert_succeed(
-        succeed!(identity).keep(zero_or_more_until(token("a"), token("b"))),
+        succeed!(identity).keep(zero_or_more_until(
+            "a letter `a`",
+            token("a"),
+            "an ending `b`",
+            token("b"),
+        )),
         "aaab",
         vec!["a", "a", "a"],
     );
     assert_succeed(
-        succeed!(identity).keep(zero_or_more_until(token("a"), token("b"))),
+        succeed!(identity).keep(zero_or_more_until(
+            "a letter `a`",
+            token("a"),
+            "an ending `b`",
+            token("b"),
+        )),
         "b",
         vec![],
     );
     assert_fail(
-        succeed!(identity).keep(zero_or_more_until(token("a"), token("b"))),
+        succeed!(identity).keep(zero_or_more_until(
+            "a letter `a`",
+            token("a"),
+            "an ending `b`",
+            token("b"),
+        )),
         "aaac",
-        "I'm expecting either the intended string or the end delimiter. However, neither was found.",
+        "I'm expecting either a letter `a` or an ending `b`. However, neither was found.",
     );
     assert_succeed(
-        succeed!(identity).keep(zero_or_more_until(token("a"), token("b"))),
+        succeed!(identity).keep(zero_or_more_until(
+            "a letter `a`",
+            token("a"),
+            "an ending `b`",
+            token("b"),
+        )),
         "aaa",
         vec!["a", "a", "a"],
     );
     assert_succeed(
-        succeed!(identity).keep(zero_or_more_until(token("a"), token("b"))),
+        succeed!(identity).keep(zero_or_more_until(
+            "a letter `a`",
+            token("a"),
+            "an ending `b`",
+            token("b"),
+        )),
         "",
         vec![],
     );
 
     assert_succeed(
         succeed!(identity).keep(zero_or_more_until(
+            "a line",
             succeed!(|line| line)
                 .keep(take_chomped(chomp_while0c(|c| *c != '\n', "line")))
                 .skip(token("\n")),
+            "an end tag",
             token("<end>"),
         )),
         "this is the 1st line
@@ -578,9 +890,11 @@ this is the 4th line
 
     assert_succeed(
         succeed!(identity).keep(zero_or_more_until(
+            "a line",
             succeed!(|line| line)
                 .keep(take_chomped(chomp_while0c(|c| *c != '\n', "line")))
                 .skip(token("\n")),
+            "an end tag",
             token("<end>"),
         )),
         "this is the 1st line
@@ -598,9 +912,11 @@ this is the 4th line
 
     assert_succeed(
         succeed!(identity).keep(zero_or_more_until(
+            "a line",
             succeed!(|line| line)
                 .keep(take_chomped(chomp_while0c(|c| *c != '\n', "line")))
                 .skip(token("\n")),
+            "an end tag",
             token("<end>"),
         )),
         "<end>",
@@ -610,14 +926,115 @@ this is the 4th line
     // edge case: the language of parser may be a superset of end_parser
     assert_succeed(
         succeed!(identity).keep(zero_or_more_until(
+            "a line",
             succeed!(|line| line)
                 .keep(take_chomped(chomp_while0c(|c| *c != '\n', "line")))
                 .skip(token("\n")),
+            "an end tag",
             token("<end>"),
         )),
         "<end>\n",
         vec![],
     );
+
+    // adapted from test_zero_or_more()
+
+    assert_fail(
+        succeed!(identity).keep(one_of!(
+            succeed!(identity)
+                .keep(zero_or_more_until(
+                    "a letter `a`",
+                    token("a"),
+                    "an end symbol",
+                    token("^")
+                ))
+                .skip(token("b")),
+            token("aab").map(|s| vec![s])
+        )),
+        "aab",
+        "I'm expecting either a letter `a` or an end symbol. However, neither was found.",
+    );
+
+    assert_fail(
+        succeed!(identity).keep(one_of!(
+            succeed!(identity)
+                .keep(zero_or_more_until(
+                    "a letter `a`",
+                    token("a"),
+                    "an end symbol",
+                    token("^")
+                ))
+                .skip(token("b")),
+            token("aab").map(|s| vec![s])
+        )),
+        "b",
+        "I'm expecting a `aab` but found `b`.",
+    );
+
+    assert_fail(
+        succeed!(identity).keep(one_of!(
+            succeed!(identity)
+                .keep(zero_or_more_until(
+                    "an `aa`",
+                    token("aa"),
+                    "an end symbol",
+                    token("^")
+                ))
+                .skip(token("!")),
+            token("aab").map(|s| vec![s])
+        )),
+        "aab",
+        "I'm expecting either an `aa` or an end symbol. However, neither was found.",
+    );
+
+    assert_fail(
+        succeed!(identity).keep(one_of!(
+            succeed!(identity)
+                .keep(zero_or_more_until(
+                    "an `aa!`",
+                    succeed!(identity).keep(token("aa")).skip(token("!")),
+                    "an end symbol",
+                    token("^")
+                ))
+                .skip(token("?")),
+            token("aab").map(|s| vec![s])
+        )),
+        "aab",
+        "I'm expecting a `!` but found `b`.",
+    );
+
+    // committed err when running end_parser
+    assert_fail(
+        succeed!(identity).keep(one_of!(
+            succeed!(identity)
+                .keep(zero_or_more_until(
+                    "an `aa`",
+                    token("aa"),
+                    "an end symbol",
+                    succeed!(identity).keep(token("b")).skip(token("!"))
+                ))
+                .skip(token("?")),
+            token("aab").map(|s| vec![s])
+        )),
+        "aab",
+        "I'm expecting a `!` but found nothing.",
+    );
+
+    assert_fail(
+        succeed!(identity).keep(one_of!(
+            succeed!(identity)
+                .keep(zero_or_more_until(
+                    "an `aa`",
+                    token("aa"),
+                    "an end symbol",
+                    succeed!(identity).keep(token("b")).skip(token("!"))
+                ))
+                .skip(token("?")),
+            token("aab").map(|s| vec![s])
+        )),
+        "b",
+        "I'm expecting a `!` but found nothing.",
+    );
 }
 
 #[test]
@@ -657,7 +1074,7 @@ fn test_any_char() {
         ParseResult::Ok {
             input: "",
             location: Location { row: 1, col: 3 },
-            output: "我",
+            output: '我',
             state: (),
             committed: false,
         }
